#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <array>
#include <vector>
#include <dirent.h>
#ifdef __APPLE__
#include <GL/glew.h>
#else
#include <OpenGL/glew.h>
#endif
#include <GLUT/glut.h>
#include "shapes.hpp"
#include <glm/glm.hpp>
#include <glm/ext.hpp>
#include <glm/gtc/type_ptr.hpp>
#include <assimp/Importer.hpp>
#include <assimp/scene.h>
#include <assimp/postprocess.h>
#include "model.hpp"
#include "program.hpp"
#include "skybox.hpp"
#include "image.hpp"
#include "util.hpp"
#include "ik.hpp"
#include "blendshapes.hpp"
#include "ui.hpp"

#pragma clang diagnostic ignored "-Wdeprecated-declarations"

using namespace std;

GLuint lightVao, cursorVao;
GLuint cursorNumIndices;

Program *textureProg, *plainProg, *reflectProg, *pbrProg, *cursorProg;

ControlWindow controlWindow;

std::vector<Skybox> skyboxes;
int activeSkybox = 0;

Assimp::Importer importer; // Need to keep this around, otherwise stuff disappears!
Model *sceneModel;

glm::vec3 camPos = {0, 0, -5}, camFront = {0, 0, 1}, camUp = {0, 1, 0};
float fov = glm::radians(30.f), znear = 0.01f, zfar = 10000.f;
float yaw = 1.57, pitch = 0;

Model *targetModel; // The model that the selection is happening on
Model::VertexLookup closestVertex;
// How close a vertex needs to be to the cursor before it is "over"
const float closestVertexThreshold = 0.5;

std::map<VertIdx, glm::vec3> manipulators;
VertIdx curManipulator = {-1,-1};

BlendshapeModel bsModel;
bool playBlendshapeAnim = false;

struct Light {
	glm::mat4 trans;
	glm::vec3 color;
};

std::vector<Light> lights;

bool discoLights = false;

int windowWidth = 800, windowHeight = 600;

enum Mode {
	Default,
	Blendshapes
};
Mode curMode;

float aspect() {
	return (float)windowWidth / (float)windowHeight;
}	

inline glm::mat4 projMat() {
	return glm::perspective(fov, aspect(), znear, zfar);
}

inline glm::mat4 viewMat() {
	return glm::lookAt(camPos, camPos + camFront, camUp);
}

void setProjectionAndViewUniforms(GLuint progId) {
	GLuint projLoc = glGetUniformLocation(progId, "projection");
	glm::mat4 proj = projMat();
	glUniformMatrix4fv(projLoc, 1, GL_FALSE, glm::value_ptr(proj));

	GLuint viewLoc = glGetUniformLocation(progId, "view");
	glm::mat4 view = viewMat();
	glUniformMatrix4fv(viewLoc, 1, GL_FALSE, glm::value_ptr(view));

	GLuint camPosLoc = glGetUniformLocation(progId, "camPos");
	glUniform3fv(camPosLoc, 1, glm::value_ptr(camPos));
}

void setLightColorAndPos(GLuint progId, glm::vec3 lightPos, glm::vec4 lightColor) {
	GLuint lightColorLoc = glGetUniformLocation(progId, "lightColor");
	glUniform4fv(lightColorLoc, 1, glm::value_ptr(lightColor));

	GLuint lightPosLoc = glGetUniformLocation(progId, "vLightPos");
	glUniform3fv(lightPosLoc, 1, glm::value_ptr(lightPos));

	GLuint viewPosLoc = glGetUniformLocation(progId, "vViewPos");
	glUniform3fv(viewPosLoc, 1, glm::value_ptr(camPos));
}

void drawPlainProg(Program *p, GLuint vao, glm::mat4 trans, glm::vec3 color) {
  	glUseProgram(p->progId);
	glBindVertexArray(vao);
	setProjectionAndViewUniforms(p->progId);
	glm::mat4 model = glm::scale(trans, glm::vec3(0.3));
	GLuint modelLoc = glGetUniformLocation(p->progId, "model");
	glUniformMatrix4fv(modelLoc, 1, GL_FALSE, glm::value_ptr(model));

	GLuint colorLoc = glGetUniformLocation(p->progId, "color");
	glUniform4fv(colorLoc, 1, glm::value_ptr(color));
		
	glDrawArrays(GL_TRIANGLES, 0, 36);
}

void drawLight(Light &light) {
	drawPlainProg(plainProg, lightVao, light.trans, light.color);
}

int findNodeTrans(const struct aiNode *n, const struct aiString name, glm::mat4 *dest) {
	if (strcmp(n->mName.data, name.data) == 0) {
		*dest = aiMatrixToMat4(n->mTransformation);
		return 0;
	}
	for (int i = 0; i < n->mNumChildren; i++) {
		if (findNodeTrans(n->mChildren[i], name, dest) == 0) {
			glm::mat4 t = aiMatrixToMat4(n->mTransformation);
			*dest = t * *dest;
			return 0;
		}
	}
	return 1;
}

glm::mat4 worldSpaceToModelSpace(aiNode *node, glm::mat4 m) {
	aiNode *parent = node;
	glm::mat4 res = m;
	std::vector<glm::mat4> trans;
	while (parent != nullptr) {
		/* res = res * glm::inverse(aiMatrixToMat4(parent->mTransformation)); */
		trans.push_back(glm::inverse(aiMatrixToMat4(parent->mTransformation)));
		parent = parent->mParent;
	}
	while (!trans.empty()) { res = trans.back() * res; trans.pop_back(); }
	return res;
}

void drawCursor(glm::vec3 pos, glm::vec3 color, float scale = 1) {
  	glUseProgram(cursorProg->progId);
	glBindVertexArray(cursorVao);
	setProjectionAndViewUniforms(cursorProg->progId);
	glm::mat4 model = glm::scale(glm::translate(glm::mat4(1), pos), glm::vec3(scale * 0.4));
	GLuint modelLoc = glGetUniformLocation(cursorProg->progId, "model");
	glUniformMatrix4fv(modelLoc, 1, GL_FALSE, glm::value_ptr(model));

	GLuint colorLoc = glGetUniformLocation(cursorProg->progId, "color");
	glUniform4fv(colorLoc, 1, glm::value_ptr(color));
		
	glDrawElements(GL_TRIANGLES, cursorNumIndices, GL_UNSIGNED_INT, 0);
}

void display() {
	glClearColor(0.5, 0.5, 0.5, 1);
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	glViewport(0, 0, windowWidth * 2, windowHeight * 2);

	float d = (float)glutGet(GLUT_ELAPSED_TIME) * 0.001f;

	glUseProgram(getUtilProg()->progId);
	setProjectionAndViewUniforms(getUtilProg()->progId);

	glUseProgram(pbrProg->progId);
	setProjectionAndViewUniforms(pbrProg->progId);

	size_t numLights = lights.size() + (discoLights ? 3 : 0);
	glm::vec3 lightPositions[numLights], lightColors[numLights];
	for (int i = 0; i < lights.size(); i++) {
		lightPositions[i] = glm::vec3(lights[i].trans[3]);
		lightColors[i] = lights[i].color;
	}

	if (discoLights) {
		for (int i = numLights - 3; i < numLights; i++) {
			auto m = glm::translate(glm::mat4(1.f), glm::vec3(-2.5, 0, 0));
			m = glm::rotate(m, glm::radians(d * 100 + i * 30), glm::vec3(1, 0, 0));
			m = glm::rotate(m, glm::radians(d * 100 + i * 30), glm::vec3(0, 1, 0));
			m = glm::rotate(m, glm::radians(d * 100 + i * 30), glm::vec3(0, 0, 1));
			lightPositions[i] = glm::vec3(m * glm::vec4(5, 0, 0, 1));
			lightColors[i] = glm::vec3(0.2);
			if (i % 3 == 0) lightColors[i].x = sin(d);
			if (i % 3 == 1) lightColors[i].y = cos(d * 3);
			if (i % 3 == 2) lightColors[i].z = cos(d);
		}
	}

	glUniform1ui(glGetUniformLocation(pbrProg->progId, "numLights"), numLights);
	glUniform3fv(glGetUniformLocation(pbrProg->progId, "lightPositions"), numLights, glm::value_ptr(lightPositions[0]));
	glUniform3fv(glGetUniformLocation(pbrProg->progId, "lightColors"), numLights, glm::value_ptr(lightColors[0]));

#ifdef COWEDBOY_IK
	{ 
		glm::vec3 targetPos(sin(d) * 2 + 3, -2, 1);
		Light targetLight = { glm::translate(glm::mat4(1), targetPos), {0.5, 1, 1}  };
		drawLight(targetLight);
		inverseKinematics(*sceneModel->find("Shoulder.L"), *sceneModel->find("Finger.L"), targetPos);

		targetPos = { sin(d * 2) * 2 - 5, 2.5, 0 };
		targetLight = { glm::translate(glm::mat4(1), targetPos), {1, 1, 0.5}  };
		drawLight(targetLight);
		inverseKinematics(*sceneModel->find("Shoulder.R"), *sceneModel->find("Finger.R"), targetPos);
	}
#endif

	if (curMode == Default)
		sceneModel->draw(skyboxes[activeSkybox], d * 1000);

	if (curMode == Blendshapes) {
		if (closestVertex.distance < closestVertexThreshold)
			drawCursor(closestVertex.pos, {0.5,0,1}, 0.9);

		for (auto v: manipulators) {
			glm::vec3 color = { 0.4, 1, 0 };
			if (closestVertex.meshIdx == v.first.first &&
				closestVertex.vertIdx == v.first.second)
				color = {1, 0, 0};
			drawCursor(v.second, color);

			glm::vec3 origVertex = aiVector3DToVec3(bsModel.model->meshes[v.first.first].ai.mVertices[v.first.second]);
			drawCursor(origVertex, {0,0,1}, 0.7);
		}

		bsModel.model->draw(skyboxes[activeSkybox], d * 1000);
	}

	for (Light &light: lights) drawLight(light);

	// TODO: restore
	/* if (discoLights) { */
	/* 	for (int i = numLights - 3; i < numLights; i++) { */
	/* 		Light l = { lightPositions[i], lightColors[i] }; */
	/* 		drawLight(l); */
	/* 	} */
	/* } */

	skyboxes[activeSkybox].draw(projMat(), viewMat());

	glutSwapBuffers();
}

void setupPlainBuffers(GLuint progId, std::vector<glm::vec3> vertices) {
	GLuint verticesSize = vertices.size() * sizeof(glm::vec3);

	glGenVertexArrays(1, &lightVao);
	GLuint vbo;
	glBindVertexArray(lightVao);
	glGenBuffers(1, &vbo);
	glBindBuffer(GL_ARRAY_BUFFER, vbo);
	glBufferData(GL_ARRAY_BUFFER, verticesSize, NULL, GL_STATIC_DRAW);
	glBufferSubData(GL_ARRAY_BUFFER, 0, verticesSize, glm::value_ptr(vertices[0]));
	GLuint posLoc = glGetAttribLocation(progId, "vPosition");
	glEnableVertexAttribArray(posLoc);
	glVertexAttribPointer(posLoc, 3, GL_FLOAT, GL_FALSE, 0, 0);
}

GLuint setupBuffersWithIndices(GLuint progId, GLuint vao,
							 std::vector<glm::vec3> vertices,
							 std::vector<GLuint> indices) {
	GLuint vbos[2];
	glBindVertexArray(vao);
	glGenBuffers(2, vbos);

	int verticesSize = vertices.size() * sizeof(glm::vec3);
	// positions
	glBindBuffer(GL_ARRAY_BUFFER, vbos[0]);
	glBufferData(GL_ARRAY_BUFFER, verticesSize, NULL, GL_STATIC_DRAW);
	glBufferSubData(GL_ARRAY_BUFFER, 0, verticesSize,
					glm::value_ptr(vertices[0]));

	GLuint posLoc = glGetAttribLocation(progId, "vPosition");
	glEnableVertexAttribArray(posLoc);
	glVertexAttribPointer(posLoc, 3, GL_FLOAT, GL_FALSE, 0, 0);

	// indices
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, vbos[1]);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, indices.size() * sizeof(GLuint), &indices[0], GL_STATIC_DRAW);

	return indices.size();
}

bool needToCalculateClosestVertex = false;
bool needToInterpolateBlendshapes = false;

class Delegate : public ControlWindowDelegate {
	public:

	virtual void weightChanged(int blendshape, float weight) {
		bsModel.blendshapes[blendshape].weight = weight;
		needToInterpolateBlendshapes = true;
	}

	virtual void solveWeights(std::vector<float> &newWeights) {
		::solveWeights(&bsModel, manipulators);
		for (int i = 0; i < newWeights.size(); i++)
			newWeights[i] = bsModel.blendshapes[i].weight;
		needToInterpolateBlendshapes = true;
	}

	virtual void resetManipulators() {
		manipulators.clear();
		curManipulator = { -1, -1 };
	}

	virtual void playbackChanged(bool playing) {
		playBlendshapeAnim = playing;
	}
};

Delegate cwDelegate;


void init() {
	initUtilProg();

	plainProg = new Program("plainvertex.glsl", "plainfrag.glsl");
	glUseProgram(plainProg->progId);
	setupPlainBuffers(plainProg->progId, cube());
	plainProg->validate();

	cursorProg = new Program("plainvertex.glsl", "plainfrag.glsl");
	glUseProgram(cursorProg->progId);
	glGenVertexArrays(1, &cursorVao);
	cursorNumIndices = setupBuffersWithIndices(cursorProg->progId, cursorVao, sphere(), sphereIndices());
	cursorProg->validate();

	skyboxes.push_back(Skybox(Image("skyboxes/loft/Newport_Loft_Ref.hdr")));
	skyboxes.push_back(Skybox(Image("skyboxes/wooden_lounge_8k.hdr")));
	skyboxes.push_back(Skybox(Image("skyboxes/machine_shop_02_8k.hdr")));
	skyboxes.push_back(Skybox(Image("skyboxes/pink_sunrise_8k.hdr")));

	pbrProg = new Program("pbrvert.glsl", "pbrfrag.glsl");
	glUseProgram(pbrProg->progId);

	if (curMode == Default) {
		const std::string scenePath = "models/cowedboy.glb";
		const aiScene *scene = importer.ReadFile(
				scenePath, aiProcess_Triangulate | aiProcess_CalcTangentSpace |
				aiProcess_GenNormals | aiProcess_FlipUVs);
		if (!scene) {
			std::cerr << importer.GetErrorString() << std::endl;
			exit(1);
		}

		if (scene->mNumCameras > 0) {
			aiCamera *cam = scene->mCameras[0];
			glm::mat4 camTrans;
			if (findNodeTrans(scene->mRootNode, cam->mName, &camTrans) != 0)
				abort(); // there must be a node with the same name as camera

			camPos = {camTrans[3][0], camTrans[3][1], camTrans[3][2]};

			glm::vec3 camLookAt =
				glm::vec3(cam->mLookAt.x, cam->mLookAt.y, cam->mLookAt.z);
			camFront = camLookAt - camPos;

			camUp = glm::vec3(cam->mUp.x, cam->mUp.y, cam->mUp.z);

			fov = cam->mHorizontalFOV;
			// TODO: aspectRatio = cam->mAspect;
			znear = cam->mClipPlaneNear;
			zfar = cam->mClipPlaneFar;
		}

		for (int i = 0; i < scene->mNumLights; i++) {
			aiLight *light = scene->mLights[i];
			glm::mat4 trans;
			findNodeTrans(scene->mRootNode, light->mName, &trans);
			glm::vec3 col = {light->mColorAmbient.r, light->mColorAmbient.g,
				light->mColorAmbient.b};
			Light l = {trans, col};
			lights.push_back(l);
		}

		sceneModel = new Model(scene, *pbrProg);
	}

	if (curMode == Blendshapes) {
		loadBlendshapes("models/high-res-blendshapes/", *pbrProg, &bsModel);
		targetModel = bsModel.model;

		size_t numBlends = bsModel.blendshapes.size();
		std::vector<std::string> names(numBlends);
		for (int i = 0; i < numBlends; i++) names[i] = bsModel.blendshapes[i].name;
		controlWindow = createControlWindow(names, &cwDelegate);

		camPos = { 0, 18, 81 };
		camFront = { 0, 0, -1 };
		camUp = { 0, 1, 0 };
		zfar = 10000;
		znear = 0.1f;
	}

	glEnable(GL_DEPTH_TEST); 
	glEnable(GL_CULL_FACE); 
	// prevent edge artifacts in specular cubemaps
	glEnable(GL_TEXTURE_CUBE_MAP_SEAMLESS);

	glViewport(0, 0, windowWidth * 2, windowHeight * 2);
}

bool keyStates[256] = {false};

void keyboard(unsigned char key, int x, int y) {
	keyStates[key] = true;
	if (key == 'z')
		activeSkybox = (activeSkybox + 1) % skyboxes.size();
	if (key == 'c')
		discoLights = !discoLights;
}

void keyboardUp(unsigned char key, int x, int y) {
	keyStates[key] = false;
}

int mouseX, mouseY;

/* #define ENABLE_MOVEMENT */
void timer(int _) {
#ifdef ENABLE_MOVEMENT
	float xSpeed = 0.f, ySpeed = 0.f, zSpeed = 0.f;

#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wchar-subscripts"
	if (keyStates['w'])
		zSpeed = 0.1f;
	if (keyStates['s'])
		zSpeed = -0.1f;
	if (keyStates['a'])
		xSpeed = 0.1f;
	if (keyStates['d'])
		xSpeed = -0.1f;
	if (keyStates['q'])
		ySpeed = 0.1f;
	if (keyStates['e'])
		ySpeed = -0.1f;
#pragma clang diagnostic pop

	camPos.x += xSpeed * sin(yaw) + zSpeed * cos(yaw);
	camPos.y += ySpeed;
	camPos.z += zSpeed * sin(yaw) - xSpeed * cos(yaw);
#endif

	if (curMode == Blendshapes) {
		float xSpeed = 0, ySpeed = 0, zSpeed = 0;
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wchar-subscripts"
		if (keyStates['w'])
			zSpeed = 0.1f;
		if (keyStates['s'])
			zSpeed = -0.1f;
		if (keyStates['a'])
			xSpeed = 0.1f;
		if (keyStates['d'])
			xSpeed = -0.1f;
		if (keyStates['q'])
			ySpeed = 0.1f;
		if (keyStates['e'])
			ySpeed = -0.1f;
#pragma clang diagnostic pop

		if (playBlendshapeAnim) {
			stepBlendshapeAnim(&bsModel);
			needToInterpolateBlendshapes = true;
			std::vector<float> newWeights(bsModel.blendshapes.size());
			for (int i = 0; i < bsModel.blendshapes.size(); i++)
				newWeights[i] = bsModel.blendshapes[i].weight;
			updateWeights(&controlWindow, newWeights);
		}

		if (curManipulator.first != -1 && curManipulator.second != -1) {
			manipulators[curManipulator].x += xSpeed;
			manipulators[curManipulator].y += ySpeed;
			manipulators[curManipulator].z += zSpeed;
		}

		if (needToInterpolateBlendshapes) {
			interpolateBlendshapes(&bsModel);
			needToInterpolateBlendshapes = false;
		}

		if (needToCalculateClosestVertex) {
			GLint vpArr[4]; glGetIntegerv(GL_VIEWPORT, vpArr);
			glm::vec4 viewport(vpArr[0], vpArr[1], vpArr[2], vpArr[3]);
			glm::vec3 selectedPos = glm::unProject(glm::vec3(mouseX * 2, viewport[3] - mouseY * 2, 1), // hidpi
					viewMat(),
					projMat(),
					viewport);

			closestVertex = targetModel->closestVertex(targetModel->getRoot(), camPos, selectedPos);
			needToCalculateClosestVertex = false;
		}
	}

	glutPostRedisplay();
	glutTimerFunc(16, timer, 0);
}

int prevMouseX, prevMouseY;
bool firstMouse = true;

void motion(int x, int y) {
	if (firstMouse) {
		prevMouseX = x;
		prevMouseY = y;
		firstMouse = false;
	}
	float dx = x - prevMouseX, dy = y - prevMouseY;
	prevMouseX = x; prevMouseY = y;
	if (curMode == Blendshapes) {
		if (closestVertex.distance > closestVertexThreshold) {
			const glm::vec3 origin(0,18,0);
			const float sensitivity = 0.003f;
			auto camMat = glm::translate(glm::mat4(1), origin + camPos);
			auto rotation = glm::rotate(glm::rotate(glm::mat4(1), -dx * sensitivity, {0, 1, 0}),
					-dy * sensitivity, {1, 0, 0});
			auto rotAroundOrig = camMat * rotation * glm::translate(glm::mat4(1), origin - camPos);
			camPos = rotAroundOrig * glm::vec4(camPos, 0);
			camFront = origin - camPos; // face center
		}
		needToCalculateClosestVertex = true;
	}
}

void passiveMotion(int x, int y) {
	if (firstMouse) {
		prevMouseX = x;
		prevMouseY = y;
		firstMouse = false;
	}
	mouseX = x; mouseY = y;
	prevMouseX = x;
	prevMouseY = y;
#ifdef ENABLE_MOVEMENT


	int dx = x - prevMouseX, dy = y - prevMouseY;

	prevMouseX = x;
	prevMouseY = y;

	const float sensitivity = 0.005f;
	yaw += dx * sensitivity;
	pitch -= dy * sensitivity;

	glm::vec3 front;
	front.x = cos(pitch) * cos(yaw);
	front.y = sin(pitch);
	front.z = cos(pitch) * sin(yaw);
	camFront = glm::normalize(front);

	if (pitch < -1.57079632679 || pitch >= 1.57079632679) {
		camUp = glm::vec3(0, -1, 0);
	} else {
		camUp = glm::vec3(0, 1, 0);
	}
#endif
	if (curMode == Blendshapes)
		needToCalculateClosestVertex = true;
}

void mouse(int button, int state, int x, int y) {
	if (isPanelFocused(controlWindow))
		return;

	if (button == GLUT_LEFT_BUTTON && state == GLUT_UP) {
		if (closestVertex.distance < closestVertexThreshold) {
			VertIdx idx = { closestVertex.meshIdx, closestVertex.vertIdx };
			if (manipulators.count(idx) <= 0)
				manipulators[idx] = closestVertex.pos;
			curManipulator = idx;
		}
	}

#ifdef ENABLE_MOVEMENT
	if (button == GLUT_LEFT_BUTTON && state == GLUT_UP)
		firstMouse = true;
#endif
}

void reshape(int newWidth, int newHeight) {
	windowWidth = newWidth, windowHeight = newHeight;
}

int main(int argc, char** argv) {
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DEPTH|GLUT_DOUBLE|GLUT_RGB|GLUT_3_2_CORE_PROFILE);
	glutInitWindowSize(windowWidth, windowHeight);
	glutCreateWindow("Physically Based Rendering");
	glutDisplayFunc(display);
	glutReshapeFunc(reshape);

	glewInit();

	// TODO: parse argv
	curMode = Blendshapes;
	init();

	glutKeyboardFunc(keyboard);
	glutKeyboardUpFunc(keyboardUp);
	glutTimerFunc(16, timer, 0);
	glutMotionFunc(motion);
	glutPassiveMotionFunc(passiveMotion);
	glutMouseFunc(mouse);

	glutMainLoop();

	return 0;
}

